/* List of special processes.

   Copyright (C) 2013 Free Software Foundation
   Written by Justus Winter <4winter@informatik.uni-hamburg.de>

   This file is part of the GNU Hurd.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifndef _HURD_PROCESSES_H
#define _HURD_PROCESSES_H

#define HURD_PID_INIT		1
#define HURD_PID_STARTUP	2
#define HURD_PID_KERNEL		3
#define HURD_PID_PROC		4

#endif  /* _HURD_PROCESSES_H */
